//+------------------------------------------------------------------+
//|                                              MQLTradeManager.mqh |
//|                                                Benjamin Fotteler |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "CC0"

#define POSITION_STRATEGY_ID 12
#define ORDER_STRATEGY_ID 15
#define DEAL_STRATEGY_ID 11

class SimpleTradeManager {
private:
   string symbol;
   ulong strategyId;
   ulong maxSlippage;
   ENUM_ORDER_TYPE_FILLING fillingType;
   ENUM_ORDER_TYPE_TIME expirationType;
      
   bool isNettingMode;
   bool isHedgingMode;
   bool isHedgingAllowed;
   bool isFIFOForced;
   double point;
   
   ulong positionId;
   bool isPositionOpening;
   bool isPositionClosing;
   ulong transactionTimeStamp;
   ulong temporaryId;

   /// I) Getter
   double GetTotalPositionVolume() {
      double openVolume = 0;
      uint total = PositionsTotal();      
      for (uint i = 0; i < total; i++) {
         string positionSymbol = PositionGetSymbol(i);
         if (symbol == positionSymbol) {
            openVolume += PositionGetDouble(POSITION_VOLUME);
         }
      }
      return openVolume;
   }
   
   double GetTotalPendingOrderVolume() {
      double openVolume = 0;
      uint total = OrdersTotal();      
      for (uint i = 0; i < total; i++) {
         ulong orderTicket = OrderGetTicket(i);
         if (symbol == OrderGetString(ORDER_SYMBOL)) {
            openVolume += OrderGetDouble(ORDER_VOLUME_CURRENT);
         }
      }
      return openVolume;
   }
   
   void GetAllowedFillingTypes(bool& isFOKAllowed, bool& isIOCAllowed, bool& isReturnAllowed) {
      ENUM_SYMBOL_TRADE_EXECUTION execution = (ENUM_SYMBOL_TRADE_EXECUTION)SymbolInfoInteger(symbol, SYMBOL_TRADE_EXEMODE);
      //in request and instant mode the execution type is set by the broker (FOK for market and REQUEST for pending)
      if (execution == SYMBOL_TRADE_EXECUTION_REQUEST || execution == SYMBOL_TRADE_EXECUTION_INSTANT) {
         isFOKAllowed = isIOCAllowed = isReturnAllowed = true;
         return;
      }
      
      uint fillingMode = (uint)SymbolInfoInteger(symbol, SYMBOL_FILLING_MODE);      
      if ((fillingMode&SYMBOL_FILLING_FOK) == SYMBOL_FILLING_FOK) isFOKAllowed = true;
      else isFOKAllowed = false;
      if ((fillingMode&SYMBOL_FILLING_IOC) == SYMBOL_FILLING_IOC) isIOCAllowed = true;
      else isIOCAllowed = false;
      //in market/broker mode the execution type RETURN is not proper for market orders
      if (execution == SYMBOL_TRADE_EXECUTION_MARKET) isReturnAllowed = false;
      else isReturnAllowed = true;
   }
   
   double RoundLotSize(const double volume) {
      double volumeStep = SymbolInfoDouble(symbol, SYMBOL_VOLUME_STEP);
      if (volumeStep <= 0) return 0;
      if (MathMod(volume, volumeStep) == 0) return volume;
      //double ratio = MathFloor(volume / volumeStep) - 1;
      double ratio = (int)(volume / volumeStep);
      return (volumeStep * ratio);
   }
   
   ulong GetFIFOCompliantTicket(const ulong positionTicket) {
      if (!PositionSelectByTicket(positionTicket)) return 0;
      if (symbol != PositionGetString(POSITION_SYMBOL)) return 0;
      ENUM_POSITION_TYPE type = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
      double volume = PositionGetDouble(POSITION_VOLUME);
      
      ulong oldestFittingTicket = 0;
      long oldestFittingTime = LONG_MAX;
      for (int i = 0; i < PositionsTotal(); i++) {
         ulong ticket = PositionGetTicket(i);
         if (PositionSelectByTicket(ticket) && symbol == PositionGetString(POSITION_SYMBOL) && type == PositionGetInteger(POSITION_TYPE) && volume == PositionGetDouble(POSITION_VOLUME)) {
           long time = PositionGetInteger(POSITION_TIME_MSC);
           if (PositionGetInteger(POSITION_TIME_MSC) < oldestFittingTime) {
              oldestFittingTicket = ticket;
              oldestFittingTime = time;
           }
         }
      }
      return oldestFittingTicket;
   }
   
   /// II) Calculate
   double CalculateMaxVolume() {
      double maxVolume = SymbolInfoDouble(symbol, SYMBOL_VOLUME_MAX); //the maximal volume of a single position
      double volumeLimit = SymbolInfoDouble(symbol, SYMBOL_VOLUME_LIMIT); //the maximal volume of a symbol
      if (volumeLimit <= 0) return maxVolume;
      double positionVolume = GetTotalPositionVolume();
      double pendingVolume = GetTotalPendingOrderVolume();
      double remainingVolume = volumeLimit - positionVolume - pendingVolume;
      return MathMin(maxVolume, remainingVolume);
   }
   
   double CalculateVolume(const double price, const double stopLoss, const ENUM_ORDER_TYPE orderType, const double risk) {
      double profit = 0;
      if (!OrderCalcProfit(orderType, symbol, 1, price, stopLoss, profit)) return 0;
      if (profit >= 0) return 0;
      double freeMargin = AccountInfoDouble(ACCOUNT_MARGIN_FREE);
      if (freeMargin <= 0) return 0;
      double calculatedVolume = RoundLotSize(risk * 0.01 * freeMargin / -profit);
      return calculatedVolume;
   }
   
   /// III) Checks
   bool CheckMargin(const double price, const double volume, const ENUM_ORDER_TYPE orderType) {
      double margin = 0;
      if (!OrderCalcMargin(orderType, symbol, volume, price, margin)) return false;
      if (AccountInfoDouble(ACCOUNT_MARGIN_FREE) < margin) return false;
      return true;
   }
   
   double CheckVolume(const double volume) {
      double minVolume = SymbolInfoDouble(symbol, SYMBOL_VOLUME_MIN);
      if (volume < minVolume) return 0;
      double maxVolume = CalculateMaxVolume();
      if (volume > maxVolume) return maxVolume;
      return RoundLotSize(volume);
   }
      
   double CheckVolume(const double volume, const double price, const ENUM_ORDER_TYPE orderType) {
      double freeMargin = AccountInfoDouble(ACCOUNT_MARGIN_FREE);
      if (freeMargin <= 0) return 0;
      
      double newVolume = volume;
      MqlTradeRequest request;
      request.action = TRADE_ACTION_DEAL;
      request.symbol = symbol;
      request.volume = volume;
      request.type = orderType;
      request.price = price;
      MqlTradeCheckResult checkResult;
   
      if (!OrderCheck(request, checkResult) && checkResult.margin_free < 0.0) {
         double coeff = freeMargin / (freeMargin - checkResult.margin_free);
         double lots = NormalizeDouble(volume * coeff, 2);
         if (lots < volume) {
            double stepvol = SymbolInfoDouble(symbol, SYMBOL_VOLUME_STEP);
            //if (stepvol > 0) newVolume = stepvol * (MathFloor(lots / stepvol) - 1);
            if (stepvol > 0) newVolume = stepvol * (int)(lots / stepvol);
         }
      }
      return CheckVolume(newVolume);
   }
   
   bool CheckStopsLevel(const double closePrice, const double stopLoss, const double takeProfit, const ulong slippage = 0) {
      bool isSLChecked = false, isTPChecked = false;
      uint stopsLevel = (int)SymbolInfoInteger(symbol, SYMBOL_TRADE_STOPS_LEVEL);
      
      if (stopLoss == 0.0) isSLChecked = true;
      else if (MathAbs(closePrice - stopLoss) > (stopsLevel + slippage) * point) isSLChecked = true;
      if (takeProfit == 0.0) isTPChecked = true;
      else if (MathAbs(closePrice - takeProfit) > (stopsLevel + slippage) * point) isTPChecked = true;;
      return (isSLChecked && isTPChecked);
   }
   
   bool CheckPositionFreeze(const ulong ticket, const ulong slippage = 0) {
      if (!PositionSelectByTicket(ticket)) return true;
      if (symbol != PositionGetString(POSITION_SYMBOL)) return false;
      uint freezeLevel = (uint)SymbolInfoInteger(symbol, SYMBOL_TRADE_FREEZE_LEVEL);
      double sl = PositionGetDouble(POSITION_SL);
      double tp = PositionGetDouble(POSITION_TP);
      
      double price = 0;
      ENUM_POSITION_TYPE positionType = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
      if (positionType == POSITION_TYPE_BUY) {
         price = SymbolInfoDouble(symbol, SYMBOL_BID);
      } else /*if (positionType == POSITION_TYPE_SELL)*/ {
         price = SymbolInfoDouble(symbol, SYMBOL_ASK);
      }
      
      bool isSLFreeze = true, isTPFreeze = true;
      if (sl == 0) isSLFreeze = false;
      else isSLFreeze = (MathAbs(price - sl) <= (freezeLevel + slippage) * point);
      if (tp == 0) isTPFreeze = false;
      else isTPFreeze = (MathAbs(tp - price) <= (freezeLevel + slippage) * point);
      
      return (isSLFreeze || isTPFreeze);
   }
   
   bool CheckFillingType(const ENUM_ORDER_TYPE_FILLING orderFillingType, const ENUM_TRADE_REQUEST_ACTIONS tradeAction = TRADE_ACTION_DEAL) {
      ENUM_SYMBOL_TRADE_EXECUTION execution = (ENUM_SYMBOL_TRADE_EXECUTION)SymbolInfoInteger(symbol, SYMBOL_TRADE_EXEMODE);
      //in request and instant mode the execution type is set by the broker (FOK for market and REQUEST for pending)
      if (execution == SYMBOL_TRADE_EXECUTION_REQUEST || execution == SYMBOL_TRADE_EXECUTION_INSTANT) return true;
   
      uint fillingMode = (uint)SymbolInfoInteger(symbol, SYMBOL_FILLING_MODE);
      if (orderFillingType == ORDER_FILLING_FOK) {
         if ((fillingMode&SYMBOL_FILLING_FOK) == SYMBOL_FILLING_FOK) return true;
         else return false;
      }
      if (orderFillingType == ORDER_FILLING_IOC) {
         if ((fillingMode&SYMBOL_FILLING_IOC) == SYMBOL_FILLING_IOC) return true;
         else return false;
      }
      if (orderFillingType == ORDER_FILLING_RETURN) {
         //in market/broker mode the execution type RETURN is not proper for market orders
         if (execution == SYMBOL_TRADE_EXECUTION_MARKET && tradeAction != TRADE_ACTION_PENDING) return false;
         return true;
      }
      return false;
   }
   
   bool CheckExpirationType(const ENUM_ORDER_TYPE_TIME timeType) {
      long  flags = 0;
      if (!SymbolInfoInteger(symbol, SYMBOL_EXPIRATION_MODE, flags)) return false;
      
      if (timeType == ORDER_TIME_GTC && (flags&SYMBOL_EXPIRATION_GTC) != 0) return true;
      else if (timeType == ORDER_TIME_DAY && (flags&SYMBOL_EXPIRATION_DAY) != 0) return true;
      else if (timeType == ORDER_TIME_SPECIFIED && (flags&SYMBOL_EXPIRATION_SPECIFIED) != 0) return true;
      else if (timeType == ORDER_TIME_SPECIFIED_DAY && (flags&SYMBOL_EXPIRATION_SPECIFIED_DAY) != 0) return true;
      return false;
   }
   
   bool IsEntryAllowed(const ENUM_POSITION_TYPE type, const double volume) {
      if (isNettingMode) {
         if (PositionSelect(symbol)) return false;
      } else if (isHedgingMode) {
         bool hasLongPositions = false;
         bool hasShortPositions = false;
         bool hasPositionsWithSameVolume = false;
         
         uint total = PositionsTotal();
         for (uint i = 0; i < total; i++) {
            string positionSymbol = PositionGetSymbol(i);
            if (symbol == positionSymbol) {
               ENUM_POSITION_TYPE positionType = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
               if (positionType == POSITION_TYPE_BUY) hasLongPositions = true;
               else if (positionType == POSITION_TYPE_SELL) hasShortPositions = true;
               double positionVolume = PositionGetDouble(POSITION_VOLUME);
               if (volume == positionVolume) hasPositionsWithSameVolume = true;
            }
         }
         if (isFIFOForced || !isHedgingAllowed) {
            if (type == POSITION_TYPE_BUY && hasShortPositions) return false;
            else if (type == POSITION_TYPE_SELL && hasLongPositions) return false;
         }
         if (hasPositionsWithSameVolume && isFIFOForced) return false;
      }
      return true;
   }
   
   // IV) Orders
   bool SendRequest(const ENUM_TRADE_REQUEST_ACTIONS action, ulong const positionTicket, const double volume, const double price, const double stopLoss, const double takeProfit, const ulong maxDeviation, const ENUM_ORDER_TYPE type, MqlTradeResult& result) {   
      MqlTradeRequest request = {action, strategyId, 0, symbol, volume, price, 0, stopLoss, takeProfit, maxDeviation, type, fillingType, expirationType, 0, NULL, positionTicket, 0};
      return OrderSend(request, result);
   }
   
   bool SendOrder(const ENUM_TRADE_REQUEST_ACTIONS tradeAction, const ulong positionTicket, const ENUM_ORDER_TYPE orderType, const double volume, const double price, const double stopLoss, const double takeProfit, MqlTradeResult& result) {
      return SendRequest(tradeAction, positionTicket, volume, price, stopLoss, takeProfit, maxSlippage, orderType, result);
   }
   
   bool SendOrder(const ENUM_TRADE_REQUEST_ACTIONS tradeAction, const ulong positionTicket, const ENUM_ORDER_TYPE orderType, const double volume, const double price, const double stopLoss, const double takeProfit, const ulong maxDeviation, MqlTradeResult& result) {
      return SendRequest(tradeAction, positionTicket, volume, price, stopLoss, takeProfit, maxDeviation, orderType, result);
   }
   
   bool ModifyPosition(const ulong positionTicket, const double stopLoss, const double takeProfit, MqlTradeResult& result) {
      return SendRequest(TRADE_ACTION_SLTP, positionTicket, 0, 0, stopLoss, takeProfit, 0, 0, result);
   }
   
   bool OpenPosition(const ENUM_POSITION_TYPE type, const double risk, const double price, const double stopLoss, const double takeProfit) {
      if (positionId != 0 || isPositionOpening) return false;
      
      double ask = SymbolInfoDouble(symbol, SYMBOL_ASK);
      double bid = SymbolInfoDouble(symbol, SYMBOL_BID);
      double entryPrice = price;
      if (price == 0) {
         if (type == POSITION_TYPE_BUY) entryPrice = ask;
         else if (type == POSITION_TYPE_SELL) entryPrice = bid;
      }
      double exitPrice = 0;
      if (type == POSITION_TYPE_BUY) exitPrice = bid;
      else if (type == POSITION_TYPE_SELL) exitPrice = ask;
      
      double sl = (type == POSITION_TYPE_BUY) ? MathMin(stopLoss, exitPrice - maxSlippage * point) : MathMax(stopLoss, exitPrice + maxSlippage * point);
      if (stopLoss == 0) sl = 0;
      double tp = (type == POSITION_TYPE_BUY) ? MathMax(stopLoss, exitPrice + maxSlippage * point) : MathMin(stopLoss, exitPrice - maxSlippage * point);
      if (takeProfit == 0) tp = 0;
      //Print("StopLoss: ",stopLoss,"; new: ",sl," # TakeProfit: ",takeProfit,"; new: ",tp);
      
      ENUM_ORDER_TYPE orderType = (type == POSITION_TYPE_BUY) ? ORDER_TYPE_BUY : ORDER_TYPE_SELL;
      
      if (!CheckStopsLevel(exitPrice, sl, tp, maxSlippage)) return false;
      double volume = CalculateVolume(entryPrice, stopLoss, orderType, risk);
      volume = CheckVolume(volume);      
      volume = CheckVolume(volume, entryPrice, orderType);
      if (!CheckMargin(entryPrice, volume, orderType)) return false;
      if (!IsEntryAllowed(type, volume)) return false;
      
      MqlTradeResult result = {};
      bool success = SendOrder(TRADE_ACTION_DEAL, 0, orderType, volume, entryPrice, sl, tp, result);
      //string resultRetCode = FormatRetCode(result);
      if (!success) {
         Print("Trade failed. Retcode: ",result.retcode);
         return false;
      } else {
         Print("Trade was successfull. Retcode: ",result.retcode);
         if (result.retcode == TRADE_RETCODE_DONE || result.retcode == TRADE_RETCODE_DONE_PARTIAL) {
            temporaryId = result.order;
            isPositionOpening = true;
            transactionTimeStamp = SymbolInfoInteger(symbol, SYMBOL_TIME_MSC);
         }
         return false;
      }
   }
   
public:

   //Basic methods
   SimpleTradeManager(const string eaSymbol, const ulong eaId, const ulong eaMaxSlippage = 100, const ENUM_ORDER_TYPE_FILLING eaFillingType = ORDER_FILLING_FOK, const ENUM_ORDER_TYPE_TIME eaTimeType = ORDER_TIME_GTC) {
      this.symbol = eaSymbol;
      this.strategyId = eaId;
      this.maxSlippage = eaMaxSlippage;
      
      if (CheckFillingType(eaFillingType)) {
         this.fillingType = eaFillingType;
      } else {
         bool isFOKAllowed = false, isIOCAllowed = false, isReturnAllowed = false;
         GetAllowedFillingTypes(isFOKAllowed, isIOCAllowed, isReturnAllowed);
         if (isFOKAllowed) this.fillingType = ORDER_FILLING_FOK;
         else if (isIOCAllowed) this.fillingType = ORDER_FILLING_IOC;
         else this.fillingType = ORDER_FILLING_RETURN;
      }
      
      if (CheckExpirationType(eaTimeType)) {
         this.expirationType = eaTimeType;
      } else {
         if (CheckExpirationType(ORDER_TIME_GTC)) this.expirationType = ORDER_TIME_GTC;
         else if (CheckExpirationType(ORDER_TIME_DAY)) this.expirationType = ORDER_TIME_DAY;
         else this.expirationType = ORDER_TIME_SPECIFIED;
      }
      
      
      ENUM_ACCOUNT_MARGIN_MODE accountMarginMode = (ENUM_ACCOUNT_MARGIN_MODE)AccountInfoInteger(ACCOUNT_MARGIN_MODE);
      this.isNettingMode = (accountMarginMode == ACCOUNT_MARGIN_MODE_RETAIL_NETTING);
      this.isHedgingMode = (accountMarginMode == ACCOUNT_MARGIN_MODE_RETAIL_HEDGING);
      this.isHedgingAllowed = AccountInfoInteger(ACCOUNT_HEDGE_ALLOWED);
      this.isFIFOForced = AccountInfoInteger(ACCOUNT_FIFO_CLOSE);
      this.point = SymbolInfoDouble(symbol, SYMBOL_POINT);
      
      Reset();
   }
   
   ~SimpleTradeManager() {
   }
   
   void OnTradeTransaction(const MqlTradeTransaction &transaction, const MqlTradeRequest &request, const MqlTradeResult &result) {
      if (!isPositionOpening && !isPositionClosing) {
         if (positionId != 0 && !PositionSelectByTicket(positionId)) positionId = 0;
         return;
      }
      if (transaction.order != temporaryId) return;
      if (transaction.type != TRADE_TRANSACTION_DEAL_ADD && transaction.type != TRADE_TRANSACTION_ORDER_DELETE && transaction.type != TRADE_TRANSACTION_HISTORY_ADD) return;
      
      if (isPositionOpening) {
         if (PositionSelectByTicket(transaction.position)) {
            positionId = transaction.position;
            temporaryId = 0;
            isPositionOpening = false;
            return;
         }
      } else if (isPositionClosing) {
         if (!PositionSelectByTicket(positionId)) {
            positionId = 0;
            temporaryId = 0;
            isPositionClosing = false;
            return;
         }
      }
   }
   
   bool NeedReset(ulong timeStamp, ulong maxDeviation) {
      if (transactionTimeStamp == 0) return false;
      if (timeStamp - transactionTimeStamp > maxDeviation) return true;
      return false;
   }
   
   void Reset() {
      this.positionId = 0;
      this.isPositionOpening = false;
      this.isPositionClosing = false;
      this.transactionTimeStamp = 0;
      this.temporaryId = 0;
      uint positionsTotal = PositionsTotal();
      for (uint i = 0; i < positionsTotal; i++) {
         ulong ticket = PositionGetTicket(i);
         if (symbol == PositionGetString(POSITION_SYMBOL) && strategyId == PositionGetInteger(POSITION_STRATEGY_ID)) {
            ENUM_POSITION_TYPE type = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
            double price = PositionGetDouble(POSITION_PRICE_OPEN);
            double volume = PositionGetDouble(POSITION_VOLUME);
            positionId = ticket;
            break;
         }
      }
   }
   
   bool AreMarketOrdersAllowed() {
      long  flags = 0;
      if (!SymbolInfoInteger(symbol, SYMBOL_ORDER_MODE, flags)) return false;    
      return ((flags&SYMBOL_ORDER_MARKET) != 0);
   }
   
   void GetPossibleStops(bool& isStopLossPossible, bool& isTakeProfitPossible) {
      uint flags = (uint)SymbolInfoInteger(symbol, SYMBOL_ORDER_MODE);
            
      if ((flags&SYMBOL_ORDER_SL) != 0) isStopLossPossible = true;
      else isStopLossPossible = false;
      if ((flags&SYMBOL_ORDER_TP) != 0) isTakeProfitPossible = true;
      else isTakeProfitPossible = false;
   }
   
   void GetPositionData(double& price, double& volume, ENUM_POSITION_TYPE &type, double& stopLoss, double& takeProfit, bool& transactionAwaited) {
      if (PositionSelectByTicket(positionId)) {
         price = PositionGetDouble(POSITION_PRICE_OPEN);
         volume = PositionGetDouble(POSITION_VOLUME);
         type = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
         stopLoss = PositionGetDouble(POSITION_SL);
         takeProfit = PositionGetDouble(POSITION_TP);
         transactionAwaited = (isPositionOpening || isPositionClosing);
      } else {
         price = volume = stopLoss = takeProfit = 0;
         transactionAwaited = (isPositionOpening || isPositionClosing);
      }
   }
   
   /// Position methods
   bool OpenLongPosition(const double risk, const double stopLoss) {
      double price = SymbolInfoDouble(symbol, SYMBOL_ASK);
      return OpenPosition(POSITION_TYPE_BUY, risk, price, stopLoss, 0);
   }
   
   bool OpenLongPosition(const double risk, const double price, const double stopLoss) {
      return OpenPosition(POSITION_TYPE_BUY, risk, price, stopLoss, 0);
   }

   bool OpenShortPosition(const double risk, const double stopLoss) {
      double price = SymbolInfoDouble(symbol, SYMBOL_BID);
      return OpenPosition(POSITION_TYPE_SELL, risk, price, stopLoss, 0);
   }
   
   bool OpenShortPosition(const double risk, const double price, const double stopLoss) {
      return OpenPosition(POSITION_TYPE_SELL, risk, price, stopLoss, 0);
   }
   
   bool ClosePosition() {
      if (positionId == 0 || isPositionClosing) return false;
      if (!PositionSelectByTicket(positionId)) return false;
      if (symbol != PositionGetString(POSITION_SYMBOL)) return false;
      if (CheckPositionFreeze(positionId)) return false;
      
      ENUM_POSITION_TYPE positionType = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
      ENUM_ORDER_TYPE orderType = ORDER_TYPE_BUY;
      double exitPrice = 0;
      if (positionType == POSITION_TYPE_BUY) {
         exitPrice = SymbolInfoDouble(symbol, SYMBOL_BID);
         orderType = ORDER_TYPE_SELL;
      } else if (positionType == POSITION_TYPE_SELL) {
         exitPrice = SymbolInfoDouble(symbol, SYMBOL_ASK);
         orderType = ORDER_TYPE_BUY;
      }
            
      double volume = PositionGetDouble(POSITION_VOLUME);      
      volume = MathMin(volume, SymbolInfoDouble(symbol, SYMBOL_VOLUME_MAX)); //our position can not be greater than SYMBOL_VOLUME_MAX
      
      MqlTradeResult result = {};
      bool success = SendOrder(TRADE_ACTION_DEAL, positionId, orderType, volume, exitPrice, 0, 0, result);
      //string resultRetCode = FormatRetCode(result);
      if (!success) {
         Print("Trade failed. Retcode: ",result.retcode);
         if (result.retcode == TRADE_RETCODE_REQUOTE) {
            double difference = 0;
            if (positionType == POSITION_TYPE_BUY) {
               difference = MathAbs(exitPrice - result.bid);
               exitPrice = result.bid;
            } else if (positionType == POSITION_TYPE_SELL) {
               difference = MathAbs(exitPrice - result.ask);
               exitPrice = result.ask;
            }
            if (difference < maxSlippage) {
               ulong nextMaxSlippage = (ulong)((maxSlippage - difference) / point);
               success = SendOrder(TRADE_ACTION_DEAL, positionId, orderType, volume, exitPrice, 0, 0, nextMaxSlippage, result);
            }
         }
      }
      if (success) {
         Print("Trade was successfull. Retcode: ",result.retcode);
         if (result.retcode == TRADE_RETCODE_DONE || result.retcode == TRADE_RETCODE_DONE_PARTIAL) {
            temporaryId = result.order;
            isPositionClosing = true;
            transactionTimeStamp = SymbolInfoInteger(symbol, SYMBOL_TIME_MSC);
         }
      }
      return false;
   }
   
   bool ModifyPosition(double newStopLoss, double newTakeProfit) {
      if (positionId == 0 || isPositionClosing) return false;
      if (!PositionSelectByTicket(positionId)) return false;
      if (symbol != PositionGetString(POSITION_SYMBOL)) return false;
      if (CheckPositionFreeze(positionId)) return false;
      
      double price = 0;
      ENUM_POSITION_TYPE positionType = (ENUM_POSITION_TYPE)PositionGetInteger(POSITION_TYPE);
      if (positionType == POSITION_TYPE_BUY) price = SymbolInfoDouble(symbol, SYMBOL_BID);
      else if (positionType == POSITION_TYPE_SELL) price = SymbolInfoDouble(symbol, SYMBOL_ASK);
      if (!CheckStopsLevel(price, newStopLoss, newTakeProfit)) return false;
      
      MqlTradeResult result;
      return ModifyPosition(positionId, newStopLoss, newTakeProfit, result);
   }
   
   bool ModifyPositionStopLoss(double newStopLoss) {
      if (!PositionSelectByTicket(positionId)) return false;
      double takeProfit = PositionGetDouble(POSITION_TP);
      return ModifyPosition(newStopLoss, takeProfit);
   }
   
   bool ModifyPositionTakeProfit(double newTakeProfit) {
      if (!PositionSelectByTicket(positionId)) return false;
      double stopLoss = PositionGetDouble(POSITION_SL);
      return ModifyPosition(stopLoss, newTakeProfit);
   }
};